
#include "Python.h"

static PyObject *py_hdr_encode(PyObject *self, PyObject *args);

static PyObject *py_hdr_decode(PyObject *self, PyObject *args);

static PyObject *py_hdr_add_array(PyObject *self, PyObject *args);
